/* -*- c-file-style: "linux"; -*- */
/**********************************************************************
 * connection_table.h                                       August 2001
 * Horms                                             horms@vergenet.net
 * Manage connections
 *
 * iproxy
 * UDP - TCP proxy
 * Copyright (C) 2001  Andrew Tridgell <tridge@samba.org>
 *                     Horms <horms@vergenet.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 **********************************************************************/


#ifndef _IPROXY_CONNECTION_TABLE_H
#define _IPROXY_CONNECTION_TABLE_H

#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>

#include "connection.h"

#define MAX_CONNECTION 128      /* This is the maximum number of simultaneous
                                   connections that iproxy-server can
                                   handle */



/**********************************************************************
 * connection_table_init
 * Initialise the connection table
 * pre: dest_ip: Address used for replies over UDP portion of connections.
 *               Ignored in Unicast mode, the IP address of the peer is
 *               used instead.
 * post: Connection table is intialised
 *       Note: This will remove any knowledge of existing connections.
 * return: 0 on success
 *         -1 on error
 **********************************************************************/

int connection_table_init(const char *dest_ip);


/**********************************************************************
 * connection_table_dump
 * Dump information about each non-PURGED connection in the connection
 * table using(connection_dump)
 * pre: none
 * post: Nothing if connection_table_init hasn't been called
 *       Information about each connection is printed to stdout otherwise
 * return: none
 **********************************************************************/

void connection_table_dump(void);


/**********************************************************************
 * connection_table_get_connection
 * Find the connection associated with a packet
 * If the packet is a command packet, process it accodingly
 * If the packet is for an unknown connection, and isn't a command
 * packet, create a connection for it.
 * pre: dest_ip: Destination IP for outgoing TCP sockets
 *      dest_port: Destination Port for outgoing TCP sockets
 *      udp_peer: sockaddr_in (IP address and port) of the peer
 *                that sent the packet
 *      packet: the packet that was sent
 *      len: the length of the packet
 * post: If the packet is a command packet
 *          Process using connection_command_packet()
 *          Return
 *       Else If the connection is unknown
 *          Create a new connection
 *       Return connection
 * return: Connection number
 *         MAX_CONNECTION if connection cannot be found
 *                        or if an error occured handling a command packet
 *                        for a known connection
 *         MAX_CONNECTION+1 if an error occured handling a command packet
 *                          for an unknown connection
 *         MAX_CONNECTION+2 If a command packet was handled successfully
 *         MAX_CONNECTION+3 if it is a packet for a similar connection
 *         That is we are reciving duplicate packets, most likely
 *         because the host has multiple interfaces routed to the other
 *         host.
 **********************************************************************/

int connection_table_get_connection(char *dest_ip, unsigned short dest_port,
                struct sockaddr_in *udp_peer, struct packet *packet, 
                int length);


/**********************************************************************
 * connection_table_read_tcp
 * Read pending data from the TCP socket associated with each
 * un-PURGED connection and pass it to the corresponding UDP connection
 * pre: read_template: Read tepmlate set with the TCP sockets that
 *                     have reads pending
 * post: Data is read from the TCP socket for each connection that
 *       had data pending. 
 *       Data is written to corresponding UDP connection
 * return: 0 on success
 *         -1 on error (Actually errors are internally ignored in
 *                      the current implementation)
 **********************************************************************/

int connection_table_read_tcp(fd_set *read_template);


/**********************************************************************
 * connection_table_set_fd_start
 * Reset internal state about set fds
 * You must call this before the first time you call
 * connection_table_set_fd_next after the read_template being refered
 * changes.
 * pre: none
 * post: Internal state about set fds is reset
 * return: none
 **********************************************************************/

void connection_table_set_fd_start(void);


/**********************************************************************
 * connection_table_set_fd_next
 * Find the next connection whose associated tcp socket is in the
 * read_tepmlate
 * pre: read_template: read_template to reference
 * post: internal status of what connection we are up to is updated.
 * return: Next connection whose TCP socket is set in the read_template
 *         -1 if no connection is found
 **********************************************************************/

int connection_table_set_fd_next(fd_set *read_template);


/**********************************************************************
 * connection_table_set_fd_end
 * Reset internal state about set fds
 * You should call this after the last time you call
 * connection_table_set_fd_next after the read_template being refered
 * changes.
 *
 * Actually it does nothing, but is here for completeness
 *
 * pre: none
 * post: none
 * return: none
 **********************************************************************/

void connection_table_set_fd_end(void);



/**********************************************************************
 **********************************************************************
 * The functions that follow enable access by connection
 * number to their connection_table_* counterparts.
 *
 * So connection_table_resend_XXX is a wrapper for
 * connection_table_XXX.
 *
 * Please see connection.h for information on the esmantics of each
 * function
 **********************************************************************
 **********************************************************************/


int connection_table_resend_packet(int con);

int connection_table_send_reply(int con, struct packet *packet);

int connection_table_send_end(int con);

int connection_table_send_ack(int con);

int connection_table_close(int con);

int connection_table_purge(int con);

int connection_table_timeout(int con);

int connection_table_read_template(fd_set *read_template);

size_t connection_table_write_all_tcp(int con, void *buf, size_t len);

uint32 connection_table_get_status(int con);

uint32 connection_table_get_tcp_fd(int con);

uint32 connection_table_get_in_offset(int con);

uint32 connection_table_get_out_offset(int con);

time_t connection_table_get_expire(int con);

time_t connection_table_get_resend(int con);

int connection_table_packet_header(int con, struct packet *packet, uint32 flag, 
                uint32 length, uint32 zero_packet);

#endif /* _IPROXY_CONNECTION_TABLE_H */
